#!/bin/sh -e

APP_NAME='klnagent'
KPD_FILENAME='klnagent.kpd'

###
# Error codes
#       1: script errors
#       2: bad values in KPD-file
#       3: package installation failed
###

### functions

read_kpd_value()
{
    local section="$1"
    local name="$2"

    awk -v "section=$section" -v "name=$name" 'BEGIN { sec = "" }
    {
        if ( $0 ~ /^\s*\[.*\]/ ) { gsub(/(^[ \t\r\n]*\[|\][ \t\r\n]*)/,""); sec=$1; next; }
        if ( ($0 ~ "^[ \t\r\n]*" name "=") && sec == section ) { gsub("(^[ \t]*" name "=|[ \t\r\n]*$)",""); print $1; }
    }' "$kpd_file" </dev/null
}

fatal_error()
{
    local err="$1" ; shift

    report "$@"

    if [ -f "$install_log" ] ; then
        cat "$install_log"
    else
        echo "$@"
    fi

    exit $err
}

report()
{
    local date="$(date "+%Y/%m/%d %T")"
    echo "[$date]" "$@" >>"$install_log"
}

run2log()
{

    report "Executing" "$@"

    "$@" >>"$install_log" 2>&1
    return $?
}

exec2log()
{
    local err="$1" ; shift

    report "Executing" "$@"

    if ! "$@" >>"$install_log" 2>&1
    then
        fatal_error $err "Installation aborted!"
    fi
    return 0
}


###

# save full path to source files
cd "$(dirname "$0")"
srcdir="$(pwd)"

# set filename for log
logdir="/var/log/kaspersky/klnagent"
install_log="$logdir/akinstall-$APP_NAME.$(date "+%Y%m%d%H%M%S").log"
if [ ! -d "$logdir" ] ; then
    mkdir "$logdir"
    chmod 0770 "$logdir"
fi

# set common variables
kpd_file="$srcdir/$KPD_FILENAME"

if [ ! -f "$kpd_file" ] ; then
    fatal_error 1 "Cannot find file $KPD_FILENAME"
fi

# get package version
version=$(read_kpd_value Version DisplayVersion)
appversion=${version%.*}
apprelease=${version##*.}

report "+++ Installation $APP_NAME version $version started"

# get type of package
pkgtype=$(read_kpd_value Setup PkgMan)
case "$pkgtype" in
    RPM)
        pkgfname="${APP_NAME}-${appversion}-${apprelease}.i386.rpm"
        ;;
    DEB)
        pkgfname="${APP_NAME}_${appversion}-${apprelease}_i386.deb"
        ;;
    *)
        fatal_error 2 "Unsupported package manager type: '$pkgtype'"
        ;;
esac
pkgfile="$srcdir/$pkgfname"

report "+++ Installing from package $pkgfname"

if [ ! -f "$pkgfile" ] ; then
    fatal_error 2 "Cannot find package file '$pkgfname'"
fi

# execute package installation
case "$pkgtype" in
    RPM)
        exec2log 3 rpm -U "$pkgfname"
        ;;
    DEB)
        dpkgopts=''
        [ "$(uname -m)" = 'x86_64' ] && dpkgopts='--force-architecture'
        exec2log 3 dpkg -i $dpkgopts "$pkgfname"
        ;;
    *)
        fatal_error 2 "Unknown package manager type: '$pkgtype'"
        ;;
esac
