#include "module.h"
#include "kavumount.h"
#include "redirfs/redirfs.h"
#include "redirfs/rfs.h"
#include <linux/security.h>
#include <linux/sched.h>
#include <linux/mount.h>

#define DEBUGFS_MAGIC		0x64626720
#define PIPEFS_MAGIC		0x50495045
#define SYSFS_MAGIC		0x62656572
#define DEVPTS_SUPER_MAGIC	0x1cd1
#define CIFS_MAGIC_NUMBER	0xFF534D42

#ifndef SMB_SUPER_MAGIC
#define SMB_SUPER_MAGIC         0x517B
#endif

#ifndef FUSE_CTL_SUPER_MAGIC
#define FUSE_CTL_SUPER_MAGIC	0x65735543
#endif

#ifndef SECURITYFS_MAGIC
#define SECURITYFS_MAGIC        0x73636673
#endif

#ifndef SELINUX_MAGIC
#define SELINUX_MAGIC		0xf97cff8c
#endif

#ifndef FUSE_SUPER_MAGIC
#define FUSE_SUPER_MAGIC	0x65735546
#endif

#define is_fuse_magic(magic)	((magic) == FUSE_SUPER_MAGIC)

inline int is_smb_magic(int magic)
{
    if (magic==SMB_SUPER_MAGIC || magic==CIFS_MAGIC_NUMBER) return 1;

    return 0;
}

inline int is_exclude_magic(int magic)
{
    if (magic == DEBUGFS_MAGIC || magic == PIPEFS_MAGIC || magic == SYSFS_MAGIC ||
            magic == DEVPTS_SUPER_MAGIC || magic == DEVFS_SUPER_MAGIC ||
            magic == PROC_SUPER_MAGIC || magic == FUSE_CTL_SUPER_MAGIC ||
            magic == SECURITYFS_MAGIC || magic == SELINUX_MAGIC) return 1;

    return 0;
}

static redirfs_filter klflt;

static struct redirfs_filter_info klflt_info = {
	.owner = THIS_MODULE,
	.name = "klflt",
	.priority = 500000000,
	.active = 1
};

static enum redirfs_rv klflt_open(redirfs_context context,struct redirfs_args *args);
static enum redirfs_rv klflt_release(redirfs_context context,struct redirfs_args *args);

static int add_path__(struct vfsmount *mnt, struct dentry *dentry, int flags);
static int register_filter(void);
static void unregister_filter(void);
static struct vfsmount *get_next_mnt(struct vfsmount *p, struct vfsmount *root);
static void cifs_calls_replace(struct dentry *d_root);
static void cifs_calls_restore(void);
static atomic_t stopped = ATOMIC_INIT(0);
static atomic_t has_binfmt = ATOMIC_INIT(0);

int use_binfmt = 1;

struct kavoas_mnt {
	struct vfsmount  *mnt;
	struct list_head list;
};

static LIST_HEAD(kavoas_mnt_list);
static RFS_DEFINE_MUTEX(kavoas_mnt_mutex);

static int registered = 0;
static RFS_DEFINE_MUTEX(registered_mutex);

#define MNT_INFILTER	0x01000000		// mount point is held by filter
#define MNT_NOFILTER	0x02000000		// mount point forced to be free
#define MNT_KAVMOUNT	(MNT_INFILTER | MNT_NOFILTER)

int Process_trusted(struct task_struct * tsk)
{
	file_info f_info;

	if(0 != get_exec_info(tsk, &f_info)) {
		if(NULL == tsk->parent) return 0;
		if(0 != get_current_euid()) return 0;
		if(0 != get_exec_info(tsk->parent, &f_info)) return 0;
	}
	
	return Monitor_kernel_trusted_proc(current->pid, &f_info);
}

static int req_queue_add_notification_too_long(struct dentry *dentry, char *fname_part)
{
	check_req_data_t *reqd;
	struct inode *inode;
	
	if(!dentry) return -1;

	inode = dentry->d_inode;

    reqd = (check_req_data_t *)MEM_ALLOC(sizeof(*reqd));
    if(!reqd) return -ENOMEM;

    check_req_init(reqd, NOTIFY_FILE);
    get_current_fs_root(&reqd->rootmnt, &reqd->root);

	if(fname_part) reqd->filename = fname_part;
	else {
		char *fn = MEM_ALLOC(dentry->d_name.len + 1);
		if(fn) {
			memcpy(fn, dentry->d_name.name, dentry->d_name.len);
			fn[dentry->d_name.len] = 0;
		}
		reqd->filename = fn;
	}

    reqd->file_op_type = FILE_NAME_TOO_LONG;
	if(inode) {
	    reqd->dev = kdev_t_to_nr(inode->i_sb->s_dev);
	    reqd->ino = inode->i_ino;
	} else {
		reqd->dev = 0;
		reqd->ino = 0;
	}

    check_req_queue_add(reqd);

	return 0;
}

static int req_queue_add_notification_vfsmnt(enum FileNotifyType file_type, struct inode *dir,
        struct dentry *dentry, struct vfsmount *vfsmnt)
{
    check_req_data_t *reqd;

    reqd = (check_req_data_t *)MEM_ALLOC(sizeof(*reqd));
    if(!reqd) return -ENOMEM;

    check_req_init(reqd, NOTIFY_FILE);
    get_current_fs_root(&reqd->rootmnt, &reqd->root);

    reqd->filename = get_path_by_dentry(dentry, vfsmnt, NULL, NULL);

    reqd->file_op_type = file_type;
    reqd->dev = kdev_t_to_nr(dir->i_sb->s_dev);
    reqd->ino = dir->i_ino;

    check_req_queue_add(reqd);

    return 0;
}

static void req_queue_add_notification(enum FileNotifyType file_type, struct inode *dir, struct dentry *dentry)
{
    struct dentry *mnt_root = dentry;
    struct vfsmount *vfs;

    while(!IS_ROOT(mnt_root)) mnt_root = mnt_root->d_parent;

    for(vfs = orig_rootmnt; vfs; vfs = get_next_mnt(vfs,orig_rootmnt)) {
        if(mnt_root == vfs->mnt_root) {
            req_queue_add_notification_vfsmnt(file_type, dir, dentry, vfs);
        }
    }
}

static void fill_check_file_req_fields(check_req_data_t *reqd, u_int32_t pid, u_int8_t file_op_type,
        int32_t flags, struct inode *inode)
{
    reqd->pid = pid;
    reqd->file_op_type = file_op_type;
    reqd->flags = flags;
    get_current_fs_root(&reqd->rootmnt, &reqd->root);
    reqd->dev = kdev_t_to_nr(inode->i_sb->s_dev);
    reqd->ino = inode->i_ino;
    reqd->mtime = inode->i_mtime.tv_sec;
    reqd->size = inode->i_size;
    reqd->owner_uid = inode->i_uid;
    reqd->file_mode = inode->i_mode;
    reqd->f_type = inode->i_sb->s_magic;
}

static int kavoas_before_open(struct inode *inode, struct file *file)
{
	file_info	   f_info;
	int		   err = 0;
    check_req_data_t *reqd;
    char *filename, *fname_part;
    
	if (!atomic_read(&monitor_started) || Process_trusted(current)
            || !file_need_check(file->f_dentry)) {
		return 0;
	}

	if(file_unlinked(file->f_dentry)) return -ENOENT;

    filename = get_path_by_dentry(file->f_dentry, file->f_vfsmnt, &err, &fname_part);
	if(0 != err) {
		if(-ENAMETOOLONG == err) {
			if(0 != req_queue_add_notification_too_long(file->f_dentry, fname_part) && fname_part) {
				MEM_FREE(fname_part);
			}
			if(!m_config.deny_too_long_fname) err = 0;
		}
		return err;
	}

	if(!file->f_dentry->d_inode->i_size) {
		MEM_FREE(filename);
		return 0;
	}

    reqd = (check_req_data_t *)MEM_ALLOC(sizeof(*reqd));
    if(!reqd) {
        MEM_FREE(filename);
        return -ENOMEM;
    }

	fill_file_info(file->f_dentry, &f_info);
    fill_check_req_common_fields(reqd, get_current_euid(), get_current_egid(), CHECK_FILE, filename);
    fill_check_file_req_fields(reqd, current->pid, FILE_OPEN_OPER, 0, file->f_dentry->d_inode); 

	err = Monitor_kernel_check_file(&f_info, reqd, &err);

    check_req_queue_release(reqd);
				
    return err;
}
enum redirfs_rv klflt_open(redirfs_context context, struct redirfs_args *args)
{
	args->rv.rv_int = kavoas_before_open(args->args.f_open.inode, args->args.f_open.file);

	return args->rv.rv_int?REDIRFS_STOP:REDIRFS_CONTINUE;
}

static int kavoas_after_release(struct inode *inode, struct file *file)
{
	file_info	   f_info;
	int		   err = 0;
    check_req_data_t *reqd;
    char *filename;

	if (!KAVMON_IS_WRITTEN(file->f_flags) ||
		    !atomic_read(&file->f_dentry->d_inode->i_writecount)) return err;

	if (!atomic_read(&monitor_started) || Process_trusted(current)
            || !file_need_check(file->f_dentry) || !file->f_dentry->d_inode->i_size) {
		 return 0;
	}

	if(file_unlinked(file->f_dentry)) return 0;

    filename = get_path_by_dentry(file->f_dentry, file->f_vfsmnt, &err, NULL);
    if(0 != err) goto l_exit;

    reqd = (check_req_data_t *)MEM_ALLOC(sizeof(*reqd));
    if(!reqd) {
        MEM_FREE(filename);
        return -ENOMEM;
    }

	fill_file_info(file->f_dentry,&f_info);
    fill_check_req_common_fields(reqd, get_current_euid(), get_current_egid(), CHECK_FILE, filename);
    fill_check_file_req_fields(reqd, current->pid, FILE_CLOSE_OPER, 0, file->f_dentry->d_inode); 

	err = Monitor_kernel_check_file(&f_info, reqd, &err);

    check_req_queue_release(reqd);

l_exit:
	return err;
}
enum redirfs_rv klflt_release(redirfs_context context,struct redirfs_args *args)
{
	int err = kavoas_after_release(args->args.f_release.inode, args->args.f_release.file);
	
	return err?REDIRFS_STOP:REDIRFS_CONTINUE;
}

static void kavoas_rename(struct inode *old_dir, struct dentry *old_dentry,
        struct inode *new_dir, struct dentry *new_dentry)
{
    struct inode *i_old, *i_new;

    if (!atomic_read(&monitor_started) || Process_trusted(current)) return;

    i_old = (old_dentry->d_inode)?old_dentry->d_inode:old_dir;
    if(inode_need_notify(i_old)) {
        req_queue_add_notification(FILE_UNLINK_OPER, i_old, old_dentry);
        Monitor_cache_del(kdev_t_to_nr(i_old->i_sb->s_dev), i_old->i_ino);
    }

    i_new = (new_dentry->d_inode)?new_dentry->d_inode:new_dir;
    if(inode_need_notify(i_new)) {
        req_queue_add_notification(FILE_RENAME_OPER, i_new, new_dentry);
        Monitor_cache_del(kdev_t_to_nr(i_new->i_sb->s_dev), i_new->i_ino);
    }
}
enum redirfs_rv klflt_rename(redirfs_context context, struct redirfs_args *args)
{
    kavoas_rename(args->args.i_rename.old_dir, args->args.i_rename.old_dentry,
        args->args.i_rename.new_dir, args->args.i_rename.new_dentry);

	return REDIRFS_CONTINUE;
}

static void kavoas_unlink(struct inode *dir, struct dentry *dentry)
{
    struct inode *inode;

	if (!atomic_read(&monitor_started) || Process_trusted(current)) return;
    
    inode = (dentry->d_inode)?dentry->d_inode:dir;
    if(inode_need_notify(inode)) {
        req_queue_add_notification(FILE_UNLINK_OPER, inode, dentry);
        Monitor_cache_del(kdev_t_to_nr(inode->i_sb->s_dev), inode->i_ino);
    }
}
enum redirfs_rv klflt_unlink(redirfs_context context, struct redirfs_args *args)
{
    kavoas_unlink(args->args.i_unlink.dir, args->args.i_unlink.dentry);

	return REDIRFS_CONTINUE;
}

enum redirfs_rv klflt_access(redirfs_context context, struct redirfs_args *args)
{
	struct inode *inode = args->args.i_permission.inode;
	if (Process_trusted(current) && args->rv.rv_int &&
			is_fuse_magic(inode->i_sb->s_magic))
		args->rv.rv_int = 0;
	return REDIRFS_CONTINUE;
}

enum redirfs_rv klflt_getattr(redirfs_context context, struct redirfs_args *args)
{
	struct dentry *dentry = args->args.i_getattr.dentry;

	if (Process_trusted(current) && args->rv.rv_int &&
			is_fuse_magic(dentry->d_sb->s_magic)) {
		generic_fillattr(dentry->d_inode, args->args.i_getattr.stat);
		args->rv.rv_int = 0;
	}

	return REDIRFS_CONTINUE;
}

static struct redirfs_op_info klflt_op_info[] = {
	{REDIRFS_REG_FOP_OPEN, klflt_open, NULL},
	{REDIRFS_REG_FOP_RELEASE, NULL, klflt_release},
	{REDIRFS_DIR_IOP_RENAME, klflt_rename, NULL},
	{REDIRFS_DIR_IOP_UNLINK, klflt_unlink, NULL},
	{REDIRFS_REG_IOP_PERMISSION, NULL, klflt_access},
	{REDIRFS_DIR_IOP_PERMISSION, NULL, klflt_access},
	{REDIRFS_LNK_IOP_PERMISSION, NULL, klflt_access},
	{REDIRFS_REG_IOP_GETATTR, NULL, klflt_getattr},
	{REDIRFS_DIR_IOP_GETATTR, NULL, klflt_getattr},
	{REDIRFS_LNK_IOP_GETATTR, NULL, klflt_getattr},
	{REDIRFS_OP_END, NULL, NULL}
};


int Monitor_intercept_init (void)
{
	return 0;

}

int Monitor_intercept_finit (void)
{

	return 0;
}

int Monitor_intercept_calls (void)
{
	DEBUG_MSG("intercepting syscalls begin");

	rfs_mutex_lock(&registered_mutex);
	atomic_set(&stopped, register_filter());
	rfs_mutex_unlock(&registered_mutex);

	return atomic_read(&stopped);
	
}

int Monitor_restore_calls (void)
{
	if (atomic_read(&stopped)) return 0;

	DEBUG_MSG("restoring syscalls begin");

	rfs_mutex_lock(&registered_mutex);
	atomic_set(&stopped, 1);
	unregister_filter();
	rfs_mutex_unlock(&registered_mutex);

	DEBUG_MSG("restoring syscalls end");
	
	return 0;
}

int rem_path_mnt(struct vfsmount *mnt)
{
    redirfs_path *paths;
    redirfs_root rmnt;
    int rv = 0;
    int i = 0;

    if(!klflt) {
        printk(KERN_ERR "klflt: NULL klflt\n");
        return -1;
    }

    rmnt = redirfs_get_root_dentry(klflt, mnt->mnt_root);
    if(!rmnt) {
        printk(KERN_ERR "klflt: get root by dentry failed\n");
        return -1;
    }

    paths = redirfs_get_paths_root(klflt, rmnt);
    if (!paths) {
        printk(KERN_ERR "klflt: NULL paths\n");
        return -1;
    }

    while (paths[i]) {
        rv = redirfs_rem_path(klflt, paths[i]);
        if (rv)
            break;
        i++;
    }

    redirfs_put_paths(paths);
    redirfs_put_root(rmnt);

    return 0;
}

static int add_path__(struct vfsmount *mnt, struct dentry *dentry, int flags)
{
	struct redirfs_path_info	flt_path_info;	
	redirfs_path			path;

	flt_path_info.dentry	= dentry;
	flt_path_info.mnt	= mnt;
	flt_path_info.flags	= flags;

	path = redirfs_add_path(klflt, &flt_path_info);
	if (IS_ERR(path)) {
		return PTR_ERR(path);
	}

	redirfs_put_path(path);
	return 0;
}

struct spoofexec {
	char const *old;
	char const *new;
};

struct spoofexec spoofexec[] = {
	{ "/bin/umount", KAVUMOUNT_PATH },
	{ NULL, NULL },
};

/* Try to find designated exec name in tree of current parents. 
 * Code will have different behaviour on RHEL5 and derivates, since
 * RHEL5 kernel uses task parent field instead of real_parent in
 * vanilla kernels.
 */
static int has_parent(char const *name)
{
	int found = 0;
	struct task_struct *parent = current->parent;

	while (parent->pid != 1) {
		char const *exe_name = get_exec_path(parent, NULL);

		if (!strcmp(exe_name, name)) found++;
		MEM_FREE(exe_name);
		parent = parent->parent;
	}
	
	return found;
}

static struct file *open_exec_nolink(char const *name)
{
	struct file *file = open_exec(name);
	char const *fpath = !IS_ERR(file) ?
		get_path_by_dentry(file->f_dentry, file->f_vfsmnt,
				NULL, NULL) : NULL;
	
	if ((!fpath || strcmp(name, fpath)) && !IS_ERR(file)) {
		fput(file);
		file = ERR_PTR(-ENOEXEC);
	}

	if (fpath) MEM_FREE(fpath);
	return file;
}

static int exec_kavoas(struct linux_binprm *bprm, struct pt_regs *regs)
{
	int retval = -ENOEXEC;
	char const *exe_name = get_path_by_dentry(bprm->file->f_dentry, bprm->file->f_vfsmnt, NULL, NULL);
	struct spoofexec *spfe = spoofexec;

	if (!exe_name) return retval;

	while (spfe->old) {
		if (!strcmp(exe_name, spfe->old) && !has_parent(spfe->new)) {
			struct file *file = open_exec_nolink(spfe->new);

			if (IS_ERR(file)) break;

			allow_write_access(bprm->file);
			fput(bprm->file);

			bprm->file = file;

			retval = prepare_binprm(bprm);
			if (retval > 0)
				retval = -ENOEXEC;

			break;
		}
		spfe++;
	}

	MEM_FREE(exe_name);

	return retval;
}

static struct linux_binfmt kavoas_binfmt = {
	.module		= THIS_MODULE,
	.load_binary	= exec_kavoas,
};

enum register_status {
        REG_NONE,
        REG_KLFLT,
        REG_OPER,
        REG_DONE,
};

int  register_filter(void)
{
	int rv = 0;

	switch (registered) {
		case REG_NONE:
			klflt = redirfs_register_filter(&klflt_info);
			if (IS_ERR(klflt)) {
				rv = PTR_ERR(klflt);
				printk(KERN_ERR "klflt: register filter failed, err=%d\n", rv);
				break;
			}

			registered++;
		case REG_KLFLT:
			rv = redirfs_set_operations(klflt, klflt_op_info);
			if (rv) {
				printk(KERN_ERR "klflt: set operations failed, err=%d\n", rv);
				break;
			}

			registered++;
		case REG_OPER:
			/* 
			 * This call puts kavoas_binfmt on top of formats list.
			 * Thus it should be called before all other binary loaders
			 * and able to spoof binary parameters data before actual
			 * loader take place.
			 *
			 * For more confidence module should periodically check
			 * whether it's binfmt_operations are on top and has not 
			 * been supressed by newly loaded binary formats. But now
			 * only need for binfmt is to spoof binaries for elf-loader.
			 * Elf-loader is permanently in-kernel and therefore could
			 * be get on top of formats.
			 */

			if (use_binfmt)
				atomic_set(&has_binfmt, insert_binfmt(&kavoas_binfmt) == 0);

			registered++;
		case REG_DONE:
			break;
		default:
			printk(KERN_ERR "klflt: bad state on init %i\n", registered);
			rv = -EINVAL;
			break;
	}

	if (rv) unregister_filter();
	return rv;
}

static void free_kmnt(struct kavoas_mnt *kmnt, int noflt)
{
	if (!noflt) {
		rem_path_mnt(kmnt->mnt);
		kmnt->mnt->mnt_flags &= ~MNT_KAVMOUNT;
	}
	
	list_del_init(&kmnt->list);
	mntput(kmnt->mnt);
	MEM_FREE(kmnt);
}

void unregister_filter(void)
{
	int rv = 0;
	struct kavoas_mnt *kmnt, *tmp;
	
        switch (registered) {
                case REG_DONE:
			if (use_binfmt && atomic_read(&has_binfmt)) {
				remove_binfmt(&kavoas_binfmt);
				atomic_set(&has_binfmt, 0);
			}

			registered--;
		case REG_OPER:
			cifs_calls_restore();

			rfs_mutex_lock(&kavoas_mnt_mutex);
			list_for_each_entry_safe(kmnt, tmp, &kavoas_mnt_list, list)
				free_kmnt(kmnt, 0);
			rfs_mutex_unlock(&kavoas_mnt_mutex);

			rv = redirfs_rem_paths(klflt);
			if (rv) {
				printk(KERN_ERR "klflt: release path failed, err=%d\n", rv);
				break;
			}

			registered--;
		case REG_KLFLT:
			rv = redirfs_unregister_filter(klflt);
			if (rv) {
				printk(KERN_ERR "klflt: unregister filter failed, err=%d\n", rv);
				break;
			} else {
				redirfs_delete_filter(klflt);
			}

			registered--;
		case REG_NONE:
			break;
		default:
			printk(KERN_ERR "klflt: bad state on exit %i\n", registered);
			rv = -EINVAL;
			break;
	}
}


struct vfsmount *get_next_mnt(struct vfsmount *p, struct vfsmount *root)
{
	struct list_head *next = p->mnt_mounts.next;
	
	spin_lock(my_vfsmount_lock);
	
	if (next == &p->mnt_mounts) {
		while (1) {
			if (p == root){
				spin_unlock(my_vfsmount_lock);
				return NULL;
			}
			next = p->mnt_child.next;
			if (next != &p->mnt_parent->mnt_mounts)
				break;
			p = p->mnt_parent;
		}
	}
	
	spin_unlock(my_vfsmount_lock);
	
	return list_entry(next, struct vfsmount, mnt_child);
}

/*
 *	kavoas_mnt_mutex should be locked by caller!
 */
int interceptor_get_path_helper(struct vfsmount *vfs, int force)
{
	if(is_smb_magic(vfs->mnt_root->d_inode->i_sb->s_magic)) {
		cifs_calls_replace(vfs->mnt_root);
		return 0;
	} else if (is_exclude_magic(vfs->mnt_root->d_inode->i_sb->s_magic)) {
		return 0;
	} else if (vfs->mnt_flags & MNT_INFILTER) {
		return 0;
	} else if (vfs->mnt_flags & MNT_NOFILTER) {
		if (force)
			vfs->mnt_flags &= ~MNT_NOFILTER;
		else
			return -ENOENT;
	}

	if (0 == add_path__(vfs, vfs->mnt_root, REDIRFS_PATH_INCLUDE)) {
		struct kavoas_mnt *kmnt = MEM_ALLOC(sizeof(struct kavoas_mnt));
		kmnt->mnt = mntget(vfs);
		kmnt->mnt->mnt_flags |= MNT_INFILTER;
		list_add_tail(&kmnt->list, &kavoas_mnt_list);
	}

	return 0;
}

static struct vfsmount *find_mnt_by_dev(char const *name, int action)
{
	struct vfsmount *vfs, *res = ERR_PTR(-ENOENT);

	if (!name) return ERR_PTR(-ENOMEM);

	for(vfs = orig_rootmnt; vfs; vfs = get_next_mnt(vfs,orig_rootmnt))
		if (!strcmp(vfs->mnt_devname, name)) {
			if (action) {
				if (!(vfs->mnt_flags & MNT_INFILTER)) {
					res = vfs;
					break;
				} else if (IS_ERR(res)) {
					res = ERR_PTR(-EBUSY);
				}
			} else {
				if (vfs->mnt_flags & MNT_INFILTER) {
					res = vfs;
				} else if (IS_ERR(res)) {
					res = ERR_PTR(-EBUSY);
				}

			}
		}
		
	return res;
}

static struct vfsmount *find_mnt_by_mnt(struct vfsmount *vfs, int action)
{
	struct vfsmount *res = ERR_PTR(-EBUSY);

	do {
		if (action) {
			if (!(vfs->mnt_flags & MNT_INFILTER)) res = vfs;
		} else {
			if (vfs->mnt_flags & MNT_INFILTER) {
				res = vfs;
				break;
			}
		}

		if (IS_ROOT(vfs->mnt_mountpoint))
			vfs = vfs->mnt_parent;
		else
			break;
	} while (!same_as_current_fs_root(vfs, vfs->mnt_root) &&
						!(vfs == vfs->mnt_parent));

	return res;
}

static struct vfsmount *interceptor_find_mnt(char const *mntfile, int action)
{
	struct nameidata nd;
	int badpath = path_lookup(mntfile, LOOKUP_FOLLOW, &nd);

	char const *name = badpath ? kstrdup(mntfile, GFP_KERNEL) :
			get_path_by_dentry(GET_NAMEIDATA_PATH(nd).dentry,
				GET_NAMEIDATA_PATH(nd).mnt, NULL, NULL);

	int isdevname = (badpath || !IS_ROOT(GET_NAMEIDATA_PATH(nd).dentry));

	struct vfsmount *vfs = isdevname ? find_mnt_by_dev(mntfile, action) :
				find_mnt_by_mnt(GET_NAMEIDATA_PATH(nd).mnt, action);

	if (IS_ERR(vfs) && isdevname && strcmp(mntfile, name))
		vfs = find_mnt_by_dev(name, action);

	if (name) MEM_FREE(name);
	if (!badpath) path_release(&GET_NAMEIDATA_PATH(nd));

	return vfs;
}

int interceptor_get_path(char const *mntfile)
{
	struct vfsmount *vfs = interceptor_find_mnt(mntfile, 1);

	if (!IS_ERR(vfs)) {
		rfs_mutex_lock(&kavoas_mnt_mutex);
		interceptor_get_path_helper(vfs, 1);
		rfs_mutex_unlock(&kavoas_mnt_mutex);
	} else {
		return PTR_ERR(vfs);
	}

	return 0;
}

int interceptor_put_path(char const *mntfile)
{
	struct vfsmount *vfs = interceptor_find_mnt(mntfile, 0);

	if (!IS_ERR(vfs)) {
		struct kavoas_mnt *kmnt, *tmp;
		int found = 0;

		if (vfs->mnt_flags & MNT_NOFILTER)
			BUG();

		rfs_mutex_lock(&kavoas_mnt_mutex);
		list_for_each_entry_safe_reverse(kmnt, tmp, &kavoas_mnt_list, list)
			if (kmnt->mnt == vfs) free_kmnt(kmnt, found++);
		vfs->mnt_flags |= MNT_NOFILTER;
		rfs_mutex_unlock(&kavoas_mnt_mutex);
	} else {
		return PTR_ERR(vfs);
	}

	return 0;
}

void Update_Include_Dirs(void)
{
	struct vfsmount * vfs;
	struct kavoas_mnt *kmnt, *tmp;

	if (atomic_read(&stopped) || IS_ERR(klflt)) return;

	rfs_mutex_lock(&kavoas_mnt_mutex);

	list_for_each_entry_safe(kmnt, tmp, &kavoas_mnt_list, list)
		if (list_empty(&kmnt->mnt->mnt_list))
			free_kmnt(kmnt, 0);

	for(vfs = orig_rootmnt; vfs; vfs = get_next_mnt(vfs,orig_rootmnt))
		interceptor_get_path_helper(vfs, 0);

	rfs_mutex_unlock(&kavoas_mnt_mutex);
}

/*
 * CIFS interceptor. It works correct only if cifs compiled as a module.
 */

#ifdef __CONFIG_CIFS_MODULE

#if 0 /* cifs interceptor dbg messages */
#define CIFS_DBGPRN(format, args...) printk(format, ##args)
#else /* dbg off */
#define CIFS_DBGPRN(format, args...)
#endif

typedef void (*pfunc_t)(void);

struct call_restore_item {
    pfunc_t *loc;   /* location of the call that is replaced */
    pfunc_t old;    /* old value of the replaced function pointer */
};

#define CR_ARR_SIZE  16
static struct call_restore_item calls_restore_array[CR_ARR_SIZE];
static int calls_restore_count = 0;
static RFS_DEFINE_MUTEX(calls_restore_mutex);
static struct module *cifs_module = NULL;

static void call_replace_nolock(pfunc_t *loc, pfunc_t new_call)
{
    if(calls_restore_count >= CR_ARR_SIZE) {
        printk(KERN_ERR "klflt: increase calls_restore_array size to properly intercept CIFS\n");
        return;
    }
    calls_restore_array[calls_restore_count].loc = loc;
    calls_restore_array[calls_restore_count].old = *loc;
    *loc = new_call;
    ++calls_restore_count;
}

typedef int (*fop_open_t)(struct inode *inode, struct file *file);
static volatile fop_open_t cifs_open = NULL;
static int smbcheck_open(struct inode *inode, struct file *file)
{
    int res;
    fop_open_t lf_open = cifs_open;

    if(!lf_open) return -ERESTARTSYS;

    res = kavoas_before_open(inode, file);

    if(0 == res) res = lf_open(inode, file);
    CIFS_DBGPRN("smbcheck_open: %s res:%d\n", file->f_dentry->d_name.name, res);

    return res;
}

typedef int (*fop_release_t)(struct inode *inode, struct file *file);
static volatile fop_release_t cifs_close = NULL;
static int smbcheck_release(struct inode *inode, struct file *file)
{
    int res; 
    fop_release_t lf_close = cifs_close;

    if(!lf_close) return -ERESTARTSYS;

    res = lf_close(inode, file);

    CIFS_DBGPRN("smbcheck_release: %s\n", file->f_dentry->d_name.name);
    kavoas_after_release(inode, file);

    return res;
}

typedef int (*iop_rename_t)(struct inode *i_src, struct dentry *d_src,
    struct inode *i_tgt, struct dentry *d_tgt);
static volatile iop_rename_t cifs_rename = NULL;
static int smbcheck_rename(struct inode *i_src, struct dentry *d_src,
    struct inode *i_tgt, struct dentry *d_tgt)
{
    int res;
    iop_rename_t lf_rename = cifs_rename;

    if(!lf_rename) return -ERESTARTSYS;

    CIFS_DBGPRN("smbcheck_rename: src:%s tgt:%s\n", d_src->d_name.name, d_tgt->d_name.name);
    kavoas_rename(i_src, d_src, i_tgt, d_tgt);

    res = lf_rename(i_src, d_src, i_tgt, d_tgt);

    return res;
}

typedef int (*iop_unlink_t)(struct inode *, struct dentry *);
static volatile iop_unlink_t cifs_unlink = NULL;
static int smbcheck_unlink(struct inode *inode, struct dentry *direntry)
{
    int res;
    iop_unlink_t lf_unlink = cifs_unlink;

    if(!lf_unlink) return -ERESTARTSYS;

    CIFS_DBGPRN("smbcheck_unlink: direntry:%s\n", direntry->d_name.name);
    kavoas_unlink(inode, direntry);

    res = lf_unlink(inode, direntry);

    return res;
}

static void replace_file_ops_nolock(struct inode *fi)
{
    if(fi->i_fop->open && smbcheck_open != fi->i_fop->open) {
        CIFS_DBGPRN("replace cifs_open\n");
        if(cifs_open && cifs_open != fi->i_fop->open) printk(KERN_ERR "klflt: different cifs_open functions\n");
        cifs_open = fi->i_fop->open;
        call_replace_nolock((pfunc_t *)&fi->i_fop->open, (pfunc_t)smbcheck_open);
    }
    if(fi->i_fop->release && smbcheck_release != fi->i_fop->release) {
        CIFS_DBGPRN("replace cifs_close\n");
        if(cifs_close && cifs_close != fi->i_fop->release) printk(KERN_ERR "klflt: different cifs_close functions\n");
        cifs_close = fi->i_fop->release;
        call_replace_nolock((pfunc_t *)&fi->i_fop->release, (pfunc_t)smbcheck_release);
    }
    if(fi->i_op->rename && smbcheck_rename != fi->i_op->rename) {
        CIFS_DBGPRN("replace cifs_rename\n");
        if(cifs_rename && cifs_rename != fi->i_op->rename) printk(KERN_ERR "klflt: different cifs_rename functions\n");
        cifs_rename = fi->i_op->rename;
        call_replace_nolock((pfunc_t *)&fi->i_op->rename, (pfunc_t)smbcheck_rename);
    }
    if(fi->i_op->unlink && smbcheck_unlink != fi->i_op->unlink) {
        CIFS_DBGPRN("replace cifs_unlink\n");
        if(cifs_unlink && cifs_unlink != fi->i_op->unlink) printk(KERN_ERR "klflt: different cifs_unlink functions\n");
        cifs_unlink = fi->i_op->unlink;
        call_replace_nolock((pfunc_t *)&fi->i_op->unlink, (pfunc_t)smbcheck_unlink);
    }
}
static int __dir_walk_replace_file_ops_nolock(struct dentry *dir, int recursive)
{
    struct dentry *de;

    rfs_for_each_d_child(de, &dir->d_subdirs) {
        if(de->d_inode) {
            if(S_IFREG == (de->d_inode->i_mode & S_IFMT)) {
                replace_file_ops_nolock(de->d_inode);
                return 1;                
            } else if(recursive && S_IFDIR == (de->d_inode->i_mode & S_IFMT)) {
                if(__dir_walk_replace_file_ops_nolock(de, 1)) return 1;
            }
        }
    }

    return 0;
}

static void dir_walk_replace_file_ops_recursive_nolock(struct dentry *dir)
{
    spin_lock(&dcache_lock);
    __dir_walk_replace_file_ops_nolock(dir, 1);
    spin_unlock(&dcache_lock);
}
static void dir_walk_replace_file_ops(struct dentry *dir)
{
    rfs_mutex_lock(&calls_restore_mutex);
    spin_lock(&dcache_lock);
    __dir_walk_replace_file_ops_nolock(dir, 0);
    spin_unlock(&dcache_lock);
    rfs_mutex_unlock(&calls_restore_mutex);
}

typedef struct dentry * (*iop_lookup_t)(struct inode *parent_dir_inode, struct dentry *direntry,
        struct nameidata *nd);
static volatile iop_lookup_t cifs_lookup = NULL;
static struct dentry *smbcheck_lookup(struct inode *parent_dir_inode, struct dentry *direntry, struct nameidata *nd)
{
    struct dentry *res;
    iop_lookup_t lf_lookup = cifs_lookup;

    if(!lf_lookup) return ERR_PTR(-ERESTARTSYS);

    res = lf_lookup(parent_dir_inode, direntry, nd);

    if(0 == res && direntry->d_inode) {
        CIFS_DBGPRN("smbcheck_lookup: i_mode:%o iop:%x fop:%x\n", direntry->d_inode->i_mode,
                (int)direntry->d_inode->i_op, (int)direntry->d_inode->i_fop);
        if(S_IFREG == (direntry->d_inode->i_mode & S_IFMT)) {
            rfs_mutex_lock(&calls_restore_mutex);
            replace_file_ops_nolock(direntry->d_inode);
            rfs_mutex_unlock(&calls_restore_mutex);
        }
    }

    return res;
}

typedef int (*fop_readdir_t)(struct file *file, void *direntry, filldir_t filldir);
static volatile fop_readdir_t cifs_readdir = NULL;
static int smbcheck_readdir(struct file *file, void *direntry, filldir_t filldir)
{
    int res;
    fop_readdir_t lf_readdir = cifs_readdir;
    
    if(!lf_readdir) return -ERESTARTSYS;
    res = lf_readdir(file, direntry, filldir);

    dir_walk_replace_file_ops(file->f_dentry);

    return res;
}

static void cifs_calls_replace(struct dentry *d_root)
{
    struct inode *di = d_root->d_inode;

    rfs_mutex_lock(&calls_restore_mutex);

    if(!cifs_module) {
        cifs_module = di->i_sb->s_type->owner;
        if(!(cifs_module && try_module_get(cifs_module))) {
            printk(KERN_ERR "klflt: can't get cifs module\n");
            cifs_module = NULL;
            return;
        }
    }

    if(di->i_op->lookup && smbcheck_lookup != di->i_op->lookup) {
        CIFS_DBGPRN("replace dir iop cifs_lookup\n");
        if(cifs_lookup && cifs_lookup != di->i_op->lookup) printk(KERN_ERR "klflt: different cifs_lookup functions\n");
        cifs_lookup = di->i_op->lookup;
        call_replace_nolock((pfunc_t *)&di->i_op->lookup, (pfunc_t)smbcheck_lookup);
    }
    if(di->i_fop->readdir && smbcheck_readdir != di->i_fop->readdir) {
        CIFS_DBGPRN("replace dir fop cifs_readdir\n");
        if(cifs_readdir && cifs_readdir != di->i_fop->readdir) printk(KERN_ERR "klflt: different cifs_readdir functions\n");
        cifs_readdir = di->i_fop->readdir;
        call_replace_nolock((pfunc_t *)&di->i_fop->readdir, (pfunc_t)smbcheck_readdir);
    }
    if(di->i_op->rename && smbcheck_rename != di->i_op->rename) {
        CIFS_DBGPRN("replace dir iop cifs_rename\n");
        if(cifs_rename && cifs_rename != di->i_op->rename) printk(KERN_ERR "klflt: different cifs_rename functions\n");
        cifs_rename = di->i_op->rename;
        call_replace_nolock((pfunc_t *)&di->i_op->rename, (pfunc_t)smbcheck_rename);
    }
    if(di->i_op->unlink && smbcheck_unlink != di->i_op->unlink) {
        CIFS_DBGPRN("replace dir iop cifs_unlink\n");
        if(cifs_unlink && cifs_unlink != di->i_op->unlink) printk(KERN_ERR "klflt: different cifs_unlink functions\n");
        cifs_unlink = di->i_op->unlink;
        call_replace_nolock((pfunc_t *)&di->i_op->unlink, (pfunc_t)smbcheck_unlink);
    }

    dir_walk_replace_file_ops_recursive_nolock(d_root);

    rfs_mutex_unlock(&calls_restore_mutex);
}
static void cifs_calls_restore(void)
{
    int i;

    rfs_mutex_lock(&calls_restore_mutex);
    for(i = 0; i < calls_restore_count; ++i) {
        *calls_restore_array[i].loc = calls_restore_array[i].old;
    }
    calls_restore_count = 0;
    if(cifs_module) {
        module_put(cifs_module);
        cifs_module = NULL;
    }
    cifs_open = NULL;
    cifs_close = NULL;
    cifs_lookup = NULL;
    cifs_rename = NULL;
    cifs_unlink = NULL;
    cifs_readdir = NULL;
    rfs_mutex_unlock(&calls_restore_mutex);
}

#else /* CIFS is not compiled as a module so below are empty stubs */
#ifdef __CONFIG_CIFS_KERNEL
#warning "Mounted samba shares can not be checked since CIFS is built into the kernel."
#endif
static void cifs_calls_replace(struct dentry *d_root) {}
static void cifs_calls_restore(void) {}
#endif
